/* ------------------------------------------------------------------------
System       : cc65 cross-development environment
Platform     : Apple IIe 128K PRODOS 8
Program      : dhishow(C) Copyright Bill Buckels 2013,2014.
               All rights reserved.

Description  : Apple //e ProDOS 8 DHGR Slideshow Program

               DHGR formats supported: A2FC and AUX,BIN file pairs

               This program builds two executables based on conditional
               compilation. The code for either is almost exactly the
               same, but they use two different memory configurations
               based on whether they use DHGR Screen One or Screen Two.

               The dhishow variation that uses Screen One runs at $4000
               directly above Screen One at $2000 and has a good deal more
               memory than the dhishow variation that uses Screen Two which
               runs at $0803 under Screen Two (at $4000).

               Either can be launched using Oliver Schmidt's loader.system
               or from BASIC.SYSTEM with the BRUN command.

Date Written : February 2013
Revision     : 1.0 First Release (Aztec C65)
               Originally released as dpshow
Date Revised : May 2014
               2.0 Second Release (cc65) - stripped-down version.

Licence      : You may use this program for whatever you wish as long
               as you agree that Bill Buckels has no warranty or
               liability obligations whatsoever from said use.
------------------------------------------------------------------------ */

#ifndef DHGR_SCREEN_ADDRESS
#define DHGR_SCREEN_ADDRESS 0x2000
#else
#if DHGR_SCREEN_ADDRESS != 0x4000
#define DHGR_SCREEN_ADDRESS 0x2000
#endif
#endif

#ifndef __APPLE2__
#define __APPLE2__
#endif

#include <stdio.h>
#include <errno.h>
#include "_file.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>

/* http://www.unix.com/man-page/posix/3/opendir/ */
/* files in current directory */
#include <dirent.h>

#include <apple2enh.h>

/* provides base address for page1 hires scanlines  */
unsigned HB[]={
0x2000, 0x2400, 0x2800, 0x2C00, 0x3000, 0x3400, 0x3800, 0x3C00,
0x2080, 0x2480, 0x2880, 0x2C80, 0x3080, 0x3480, 0x3880, 0x3C80,
0x2100, 0x2500, 0x2900, 0x2D00, 0x3100, 0x3500, 0x3900, 0x3D00,
0x2180, 0x2580, 0x2980, 0x2D80, 0x3180, 0x3580, 0x3980, 0x3D80,
0x2200, 0x2600, 0x2A00, 0x2E00, 0x3200, 0x3600, 0x3A00, 0x3E00,
0x2280, 0x2680, 0x2A80, 0x2E80, 0x3280, 0x3680, 0x3A80, 0x3E80,
0x2300, 0x2700, 0x2B00, 0x2F00, 0x3300, 0x3700, 0x3B00, 0x3F00,
0x2380, 0x2780, 0x2B80, 0x2F80, 0x3380, 0x3780, 0x3B80, 0x3F80,
0x2028, 0x2428, 0x2828, 0x2C28, 0x3028, 0x3428, 0x3828, 0x3C28,
0x20A8, 0x24A8, 0x28A8, 0x2CA8, 0x30A8, 0x34A8, 0x38A8, 0x3CA8,
0x2128, 0x2528, 0x2928, 0x2D28, 0x3128, 0x3528, 0x3928, 0x3D28,
0x21A8, 0x25A8, 0x29A8, 0x2DA8, 0x31A8, 0x35A8, 0x39A8, 0x3DA8,
0x2228, 0x2628, 0x2A28, 0x2E28, 0x3228, 0x3628, 0x3A28, 0x3E28,
0x22A8, 0x26A8, 0x2AA8, 0x2EA8, 0x32A8, 0x36A8, 0x3AA8, 0x3EA8,
0x2328, 0x2728, 0x2B28, 0x2F28, 0x3328, 0x3728, 0x3B28, 0x3F28,
0x23A8, 0x27A8, 0x2BA8, 0x2FA8, 0x33A8, 0x37A8, 0x3BA8, 0x3FA8,
0x2050, 0x2450, 0x2850, 0x2C50, 0x3050, 0x3450, 0x3850, 0x3C50,
0x20D0, 0x24D0, 0x28D0, 0x2CD0, 0x30D0, 0x34D0, 0x38D0, 0x3CD0,
0x2150, 0x2550, 0x2950, 0x2D50, 0x3150, 0x3550, 0x3950, 0x3D50,
0x21D0, 0x25D0, 0x29D0, 0x2DD0, 0x31D0, 0x35D0, 0x39D0, 0x3DD0,
0x2250, 0x2650, 0x2A50, 0x2E50, 0x3250, 0x3650, 0x3A50, 0x3E50,
0x22D0, 0x26D0, 0x2AD0, 0x2ED0, 0x32D0, 0x36D0, 0x3AD0, 0x3ED0,
0x2350, 0x2750, 0x2B50, 0x2F50, 0x3350, 0x3750, 0x3B50, 0x3F50,
0x23D0, 0x27D0, 0x2BD0, 0x2FD0, 0x33D0, 0x37D0, 0x3BD0, 0x3FD0};


unsigned delay = 5;

/* move a block of data from main to auxiliary memory */
void maintoaux(unsigned src0, unsigned src1, unsigned dest0)
{

    unsigned *src  = (unsigned *)0x3c;
    unsigned *dest = (unsigned *)0x42;

    src[0] = src0;
    src[1] = src1;
    dest[0] = dest0;

    asm("sec");
    asm("jsr $c311");

}

/* 80 column mode must be set before calling */
#pragma optimize (push,off)
void dhireson(void)
{
    /* page 1 double hires */
    asm("sta $c000"); /* TURN OFF 80 STORE        */
    asm("sta $c050"); /* GRAPHICS                 */
#if DHGR_SCREEN_ADDRESS == 0x4000
    asm("sta $c055"); /* PAGE TWO                 */
#else
    asm("sta $c054"); /* PAGE ONE                 */
#endif
    asm("sta $c052"); /* GRAPHICS ONLY, NOT MIXED */
    asm("sta $c057"); /* HI-RES                   */
    asm("sta $c05e"); /* TURN ON DOUBLE RES       */
}
#pragma optimize (pop)

void dhiresclear(void)
{
    /* clear main screen memory */
    memset((char *)DHGR_SCREEN_ADDRESS,0,8192);
    /* clear auxiliary screen memory */
    maintoaux(DHGR_SCREEN_ADDRESS,DHGR_SCREEN_ADDRESS+8191,8192);
}

/* 80 column mode must be set to on after calling */
#pragma optimize (push,off)
void dhiresoff(void)
{
    asm("sta  $c051"); /* TEXT - HIDE GRAPHICS */
    asm("sta  $c05f"); /* TURN OFF DOUBLE RES  */
    asm("sta  $c054"); /* PAGE ONE             */
    asm("sta  $c001"); /* TURN ON 80 STORE     */
}
#pragma optimize (pop)

char lodebuf[1024];
/* The following loads two variations of BSaved
   non-compressed double hi-res graphics image formats:

   BSaved AUX, BIN Image Format file pairs
   A2FC combined file of BSaved AUX, BIN file pairs

*/
int dlodehi(char *name)
{
   FILE *fp;
   int c, fa = 0, fl = 8192, jdx, idx, status=0;
   char name2[66];
   unsigned src0, src1, dest;

   jdx = 999;
   for (idx = 0; name[idx] != 0; idx++) {
        if (name[idx] < (char)33 || name[idx] > (char)122)
            name[idx] = (char)0;
        name2[idx] = name[idx];
        if (name[idx] == '.') jdx = idx;
   }
   name2[idx] = 0;

   /* try to open a BASIC AUX, BIN image pair */
   if (jdx != 999) name2[jdx] = 0;
   strcat(name2,".AUX");

   /* is it a bsaved AUX file ? */
   fl = 8192;
   fp = fopen(name2,"rb");
   if (fp == NULL) {
       /* if not assume it's an A2FC file. */
       fl = 16384;
       fp = fopen(name,"rb");
       if (fp == NULL)return -1;
   }

   for (;;) {
       src0 = (unsigned)&lodebuf[0];
       src1 = src0 + 1023;
       dest = DHGR_SCREEN_ADDRESS;

       for (idx = 0;idx < 8; idx++) {
           /* read to main memory */
            c = fread(lodebuf,1,1024,fp);
            if (idx == 7 && fl == 8192) {
                /* AUX files can be 8192 or 8184 bytes */
                if (c < 1016) {
                    status = -2; break;

                }
            }
            else {
                if (c != 1024) {
                    status = -2; break;
                }
            }
            /* move to auxiliary screen memory */
            maintoaux(src0,src1,dest);
            dest += 1024;
       }

       if (status !=0) {
		   fclose(fp);
		   break;
       }

       if (fl == 8192) {
            fclose(fp);
            fp = fopen(name,"rb"); /* open a binary file */
            if (fp == NULL)return -1;
       }

       dest = DHGR_SCREEN_ADDRESS;
       for (idx = 0;idx < 8; idx++) {
           /* read to main memory */
            c = fread(lodebuf,1,1024,fp);
            if (idx == 7) {
                /* BIN files can be 8192 or 8184 bytes */
                if (c < 1016) {
                    status = -2; break;

                }
            }
            else {
                if (c != 1024) {
                    status = -2; break;
                }
            }
            /* move to main screen memory */
            memcpy((char *)dest,(char *)&lodebuf[0],1024);
            dest += 1024;
       }
       break;
   }

   fclose(fp);

   return status;
}


int dfraglode(char *name,int hclear)
{
	FILE *fp;
	int c,y,y1,xoff,status=-1,width,height,packet;
	unsigned src0, src1, dest;
	unsigned char *ptr;

	fp = fopen(name,"rb");
	if (fp == NULL) return -1;
	for (;;) {
		/* read 5 byte header */
		c = fread(lodebuf,1,5,fp);
		if (c != 5) break;
		if (lodebuf[0] != 'D' || lodebuf[1] != 'H' || lodebuf[2] != 'R') break;
		width = (int)lodebuf[3];
		height = (int)lodebuf[4];
		/* must be in a valid range */
		if (width < 4 || width > 80) break;
		if (height< 1 || height > 192) break;
		status = 0;

        if (hclear == 1) dhiresclear();

		packet = width / 2;
       	src0 = (unsigned)&lodebuf[0];
       	src1 = src0 + packet - 1;

        /* center fragment on screen */
        y1 = 96 - (height/2);
        xoff = 20 - (packet / 2);

		for (y = 0; y < height;y++,y1++) {
			/* read the keyboard buffer */
			/* stop if keypress */
			ptr = (unsigned char*)0xC000;
			if (ptr[0] > 127) break;

			c = fread(lodebuf,1,width,fp);
			if (c!= width) {
				status = -2;
				break;
			}
#if DHGR_SCREEN_ADDRESS == 0x4000
            dest = HB[y1] + 0x2000;
#else
			dest = HB[y1];
#endif

            dest += xoff;

            /* move to auxiliary screen memory */
	        maintoaux(src0,src1,dest);
			/* move to main screen memory */
			memcpy((char *)dest,(char *)&lodebuf[packet],packet);

		}
		break;
	}

   fclose(fp);
   return status;
}


/* 1MHZ timing loop settings - increase for faster machines */
#define YTIME 60L
#define XTIME 32L
char wait(unsigned duration)
{
    char c = 0;
    long y, x;

    while (duration > 0) {
        for (y=0;y<YTIME;y++) {
            for (x=0;x<XTIME;x++) {
                if (kbhit()) {
                    c = cgetc();
                	if (c > 47 && c < 58) {
						delay = (c-48) * 5;
					}
                    /* clear keyboard buffer */
                    while (kbhit())cgetc();
                    return c;
                }
            }
        }
        duration--;
    }
    return c;
}

/* the following function creates a list of the double hi-res
   files in the current directory, in Apple sequential text format. */
FILE *makepiclist()
{
    FILE *fp = NULL;
    int err, cnt = 0, idx, jdx, found, frag;
    DIR *dir;
    struct dirent *dp;

    extern unsigned char _filetype;
    extern unsigned _auxtype;

    /* read files in current directory */
    if ((dir = opendir (".")) == NULL) return NULL;

    /* remove piclist if it already exists */
    if((fp=fopen("PICLIST","r"))!=NULL) {
        fclose(fp);
        remove("PICLIST");
    }

    /* set file to sequential text */
    _filetype = 0x04;
    _auxtype = 0x0000;

    if((fp=fopen("PICLIST","w"))==NULL) {
        closedir(dir);
        return NULL;
    }

    /* write the list to disk */
    while ((dp = readdir (dir)) != NULL) {
        /* check for a2fc images and bin and aux image pairs */
        /* binary and FOT files only */
        found = frag = 0;

        if (dp->d_type == 0x08) {
            /* bypass all FOT files except for A2FC style files */
            if (dp->d_auxtype > 0x3fff) continue;
            if (dp->d_size != 16384) continue;
            found = 1;
        }
        else {

            if (dp->d_type != 0x06) continue;
            /* load address 0x2000 or 0x4000 */
            if (dp->d_auxtype  != 0x2000 && dp->d_auxtype != 0x4000) continue;
            /* check file size */
            /* there is no way to tell the difference between a
            hi-res bsaved image and a double hi-res bin or aux image */
            found = 0;
            if (dp->d_size > 8183  && dp->d_size < 8193) found = 1;
            if (dp->d_size > 16375 && dp->d_size < 16385)found = 1;
            if (found == 0) {
				/* image fragments */
				if (dp->d_size > 15365 || dp->d_size < 9) continue;
				frag = 1;
			}
        }

        /* get extension if any */
        jdx = 999;
        for (idx = 0; dp->d_name[idx] != 0; idx++) {
            if (dp->d_name[idx] == '.') {
                jdx = idx+1;
            }
        }

		if (frag == 1) {
			if (jdx == 999) continue;
            for (;;) {
				/* image fragments must have the DHR extension */
                if (toupper(dp->d_name[jdx])   != 'D') break;
                if (toupper(dp->d_name[jdx+1]) != 'H') break;
                if (toupper(dp->d_name[jdx+2]) != 'R') break;
                found = 1;
                break;
            }
            if (found == 0) continue;

		}
		else {
			if (jdx != 999) {
				found = 0;
				for (;;) {
					/* bypass AUX files - I only want BIN files in my list */
					if (toupper(dp->d_name[jdx])   != 'A') break;
					if (toupper(dp->d_name[jdx+1]) != 'U') break;
					if (toupper(dp->d_name[jdx+2]) != 'X') break;
					found = 1;
					break;
				}
				if (found == 1) continue;
			}
		}


        /* if out of disk space quit writing to piclist */
        if (frag == 1) err = fprintf(fp,"#%s%c",dp->d_name, (char)13);
        else err = fprintf(fp,"%s%c",dp->d_name, (char)13);
        if (err < 0) break;
        cnt++;
    }

    fclose(fp);
    closedir(dir);

    if (cnt == 0) {
        /* if no files found or no filenames written remove empty piclist */
        remove("PICLIST");
        fp = NULL;
    }
    else {
        /* otherwise open and return the piclist */
        fp = fopen("PICLIST","r");
    }

    return fp;
}

/* fgets for Apple II text files - read native apple II text files properly!
   the cc65 version doesn't! Which means that you need to use unix text files
   unless you roll your own fgets like this one which does both... */
char* __fastcall__ fgets (char* s, unsigned size, register FILE* f)
{
    register char* p = s;
    unsigned i;
    int c;

    if (size == 0) {
        /* Invalid size */
        return (char*) _seterrno (EINVAL);
    }

    /* Read input */
    i = 0;
    while (--size) {

        /* Get next character */
        if ((c = fgetc (f)) == EOF) {
            /* Error or EOF */
            if ((f->f_flags & _FERROR) != 0 || i == 0) {
                /* ERROR or EOF on first char */
                *p = '\0';
                return 0;
            } else {
                /* EOF with data already read */
                break;
            }
        }

        /* One char more */
        *p = c;
        ++p;
        ++i;

        /* Stop at end of Apple II or Unix Text File line */
        if ((char)c == (char)13 || (char)c == (char)10) {
            break;
        }
    }

    /* Terminate the string */
    *p = '\0';

    /* Done */
    return s;
}



int main(void)
{
    FILE *fp;
    int i,status = 0;
    char picname[66], c;
    unsigned cnt = 0;

    /* allow delay change from file if found */
    if((fp=fopen("delay.txt","r"))!=NULL) {
       if (fgets(picname,66,fp) != NULL) delay = (unsigned)atoi(picname);
       fclose(fp);
       printf("Delay Over-ride %d. Press any key...\n",delay);
       cgetc();
    }


    /* try to create a piclist if it does not exist */
    if((fp=fopen("PICLIST","r"))==NULL) {

        fp = makepiclist();
        if (fp == NULL) {
            printf("Can't open PICLIST! Press any key...");
            cgetc();
            return 0;
        }
    }

    /* initialize 80 column card */
    videomode(VIDEOMODE_80COL);
    clrscr();

    /* turn-on double hi-res and clear the dhgr screen */
    dhireson();
    dhiresclear();

    /* default */
    /* display each dhgr pic for 5 seconds (default) or advance on
       a key press. If ESC is pressed, exit */

    for (;;) {
        while (fgets(picname,66,fp) != NULL) {
            /* load picture */
            /* skip line-feeds in MS-DOS text */
            /* skip comment lines and blank lines if any */
            if (picname[0] < (char)47 && picname[0] != '#') continue;

            /* remove trailing eol or unwanted characters */
            for (i=0;picname[i]!=0;i++) {
				if (picname[i] < (char)32 || picname[i] > (char)122) {
					picname[i] = 0;
					break;
				}
			}
            if (picname[0] == '#') {
				if (dfraglode((char *)&picname[1],1) < 0) continue;
			}
			else {
            	if (dlodehi(picname) < 0) continue;
			}
            if (cnt++ > 100) cnt = 1;
            if (delay == 0) {
                /* if delay == 0 advance on keypress only */
                /* if ESC is pressed then quit */
                c = cgetc();
                if (c == 27) {
                    status = -1; break;
                }

                if (c > 48 && c < 58) {
					delay = (c-48) * 5;
	                while (kbhit())cgetc();
					continue;
				}

            }
            else {
            	/* flush keyboard buffer */
            	if (kbhit()) c = cgetc();
            	else c = 0;

                if (c == 27) {
                    status = -1; break;
                }

                if (c > 47 && c < 58) {
					delay = (c-48) * 5;
				}


			}
            while (kbhit())cgetc();

            if (delay == 0) continue;

            /* advance on timeout or advance on keypress */
            /* if ESC is pressed then quit */
            if (wait(delay)==27) {
                status = -1;
                break;
            }

        }
        fclose(fp);
        if (status != 0) break;

        if (cnt < 1) {
            fp = makepiclist();
            if (fp == NULL) break;
            continue;
        }

        /* no rewind in cc65, so we must
           close and re-open the piclist when we
           reach the end */
        if((fp=fopen("PICLIST","r"))==NULL) break;
    }

    dhiresclear();
    dhiresoff();
    videomode(VIDEOMODE_80COL);
    clrscr();

    if (status == 0 && cnt == 0) {
        printf("Can't create PICLIST! Press any key...");
        cgetc();
    }

    return 0;
}
